#include <mainwindow.h>
#include "ui_mainwindow.h"

#include <QFileDialog>
#include <QMessageBox>
#include <QPixmap>
#include <settingsdialog.h>
#include <string>
#include <boost/regex.hpp>

#ifdef BUILD_OFD_LOCAL_QR_SCAN
#   include <adjustpicturedialog.h>
#   include <checkqueuetablemodel.h>
#endif

#include <outputdialog.h>
#include <solvecaptchadialog.h>
#include <net/net.h>
#include <utils/utils.h>
#include <exceptions/ofdrequestexception.h>

#ifdef BUILD_OFD_BINARYEYE_SCAN
#   include <qrencode.h>
#   include <stdlib.h>
#   include <unistd.h>
#   include <sys/types.h>
#   include <sys/socket.h>
#   include <netinet/in.h>
#include <email_parser/emailparser.h>
#endif

MainWindow::MainWindow(QWidget *parent)
    : QWidget(parent)
    , ui(new Ui::MainWindow) {
    ui->setupUi(this);
    ui->stop_server_button->hide();

    ui->checks_to_parse_label->hide();
    ui->checkQueueTable->hide();
    ui->deleteSelectedButton->hide();
    ui->parse_button->hide();

    model = new CheckQueueTableModel(&checks, this);
    ui->checkQueueTable->setModel(model);
    ui->checkQueueTable->viewport()->setAcceptDrops(true);
    ui->checkQueueTable->setDragDropMode(QAbstractItemView::DragDrop);

    ui->checkQueueTable->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);

#ifdef BUILD_OFD_BINARYEYE_SCAN
    QObject::connect(this, &MainWindow::httpErrorOccured, this, &MainWindow::notifyHttpServerFailure);
    connect(this, SIGNAL(httpNewMessage(QString)), this, SLOT(httpNewMessageHandler(QString)));
#else
    ui->or_label_2->hide();
    ui->binary_eye_button->hide();
#endif

#ifndef BUILD_EMAIL_MODE
    ui->parse_email_button->hide();
    ui->or_label_2->hide();
#endif

#ifndef BUILD_OFD_LOCAL_QR_SCAN
    ui->or_label_1->hide();
    ui->choose_image_button->hide();
#endif

    QObject::connect(ui->settings_button, &QPushButton::clicked, [&]() {
        SettingsDialog d;
        d.show();
        d.exec();
    });
}

#ifdef BUILD_OFD_BINARYEYE_SCAN
void MainWindow::startHttpServer() {
    server = new HttpServer(this);

    if (server->start() < 0) {
        emit httpErrorOccured();
    }
}

void MainWindow::on_binary_eye_button_clicked() {
    httpServerThread = new std::thread(&MainWindow::startHttpServer, this);
    ui->binary_eye_button->setEnabled(false);
    ui->stop_server_button->show();

    while (!server->isStarted()) {}

    std::string localIp;
    try {
        localIp = get_local_ip_address();
    } catch(std::exception e) {
        std::cerr << e.what() << std::endl;
        return;
    }

    std::string connectionString = "binaryeye://scan?ret=http://" + localIp + ":" + std::to_string(server->getPort()) + "/?result={RESULT}";

    generate_qr_code(connectionString);

    QMessageBox infoDialog = QMessageBox();
    infoDialog.setText(QString::fromStdString(connectionString));
    infoDialog.setIconPixmap(QPixmap(QString::fromStdString(get_path_relative_to_home(".local/share/checks_parser/binaryeye_connection.png"))).scaled(400, 400, Qt::KeepAspectRatio));
    infoDialog.setWindowTitle(tr("QR code for binaryeye to connect"));
    infoDialog.addButton(tr("I've scanned"), QMessageBox::ButtonRole::AcceptRole);
    infoDialog.exec();
}

void MainWindow::notifyHttpServerFailure() {
    QMessageBox infoDialog = QMessageBox();
    infoDialog.setText(tr("Could not start http server. 10 times in a row random port was occupied. Either you should run for a lottery ticket, or the problem is in the program. If the lottery ticket wasn't lucky, please, contact the developer."));
    infoDialog.setIcon(QMessageBox::Warning);
    infoDialog.setWindowTitle(tr("Could not start http server."));
    infoDialog.exec();
}

void MainWindow::on_stop_server_button_clicked() {
    delete server;
    ui->stop_server_button->hide();
    ui->binary_eye_button->setEnabled(true);
}

void MainWindow::httpNewMessageHandler(QString message) {
    std::string parametersString = split(message.toStdString(), " ")[1];

    //erase /?result= from the string
    parametersString.erase(0, parametersString.find("=") + 1);
    std::map<std::string, std::string> paramsMap = get_params_from_string(parametersString);
    emit onDataDecode(paramsMap);
}

#endif //ifdef BUILD_OFD_BINARYEYE_SCAN

#ifdef BUILD_OFD_LOCAL_QR_SCAN
void MainWindow::on_choose_image_button_clicked() {
    QString filename = QFileDialog::getOpenFileName();

    if (filename == "") {
        QMessageBox infoDialog;
        infoDialog.setText(tr("Please, select a picture where QR code that contains info about check is present"));
        infoDialog.setIcon(QMessageBox::Critical);
        infoDialog.setWindowTitle(tr("Picture was not selected"));
        infoDialog.exec();
        return;
    }

    ui->info_label->setText(tr("Selected image: ") + filename);

    AdjustPictureDialog dialog = AdjustPictureDialog(this, filename.toStdString());
    connect(&dialog, &AdjustPictureDialog::decodedData, this, &MainWindow::onDataDecode);
    dialog.exec();
}
#endif //ifdef BUILD_OFD_LOCAL_QR_SCAN

void MainWindow::onDataDecode(std::map<std::string, std::string> data) {
    set_check_params(data);
}

#ifdef BUILD_EMAIL_MODE

void MainWindow::on_parse_email_button_clicked() {
    QString filename = QFileDialog::getOpenFileName();

    if (filename == "") {
        QMessageBox infoDialog;
        infoDialog.setText(tr("Please, select an e-mail which contains QR code"));
        infoDialog.setIcon(QMessageBox::Critical);
        infoDialog.setWindowTitle(tr("E-Mail was not selected"));
        infoDialog.exec();
        return;
    }

    EmailParser email_parser;
    std::map<std::string, std::string> paramsMap = email_parser.parse_file(filename.toStdString());

    set_check_params(paramsMap);
    // QMessageBox infoDialog;
    // infoDialog.setText(tr("This feature is under development. Wait it to appear in next updates."));
    // infoDialog.setIcon(QMessageBox::Warning);
    // infoDialog.setWindowTitle(tr("Under development"));
    // infoDialog.exec();
    // return;
}

#endif // ifdef BUILD_EMAIL_MODE

void MainWindow::on_parse_button_clicked() {

    if (checks.empty()) {
        QMessageBox infoDialog;
        infoDialog.setText(tr("Please, add check(s) to parse"));
        infoDialog.setIcon(QMessageBox::Warning);
        infoDialog.setWindowTitle(tr("No checks to parse"));
        infoDialog.exec();
        return;
    }

    OutputDialog d = OutputDialog(this, &checks);
    d.exec();
}

void MainWindow::on_add_new_check_button_clicked() {
    Check *new_check = parse_new_check();
    if (new_check == nullptr) {
        return;
    }

    unsigned int newRowIndex = checks.size();
    model->insertRows(newRowIndex, 1);

    checks.at(newRowIndex) = *new_check;

    emit model->dataChanged(model->index(newRowIndex, 0), model->index(newRowIndex, 1));

    delete new_check;

    if (checks.size() > 0) {
        ui->checkQueueTable->show();
        ui->checks_to_parse_label->show();
        ui->deleteSelectedButton->show();
        ui->parse_button->show();
    }
}

Check *MainWindow::parse_new_check() {
    Net net;
    std::string solved_captcha = "";

    for (unsigned short i = 0; i < 5; i ++) {
        // Will be repaced with neural network to solve captchas as soon as I train it.
        net.get_captcha_from_ofdru();
        SolveCaptchaDialog dialog = SolveCaptchaDialog(this, &solved_captcha);
        dialog.exec();
        Check* check = new Check();

        try {
            std::string check_content = net.fetch_check_data_from_ofdru(
                ui->fn_line_edit->text().toStdString(),
                ui->fd_line_edit->text().toStdString(),
                ui->fi_line_edit->text().toStdString(),
                ui->purchase_datetime_edit->dateTime().toString(Qt::ISODate).toStdString(),
                ui->operation_type_combo_box->currentIndex() + 1,
                // In the request to ofd.ru, total is in a format with 2 last digits represent decimal part of a number.
                ui->total_spin_box->text().toDouble() * 100,
                solved_captcha
            );

            (*check) = parseOfdRuAnswer(check_content);
            check->set_date(ui->purchase_datetime_edit->dateTime().toString(Qt::ISODate).toStdString());
            check->set_fn(ui->fn_line_edit->text().toStdString());
            check->set_fd(ui->fd_line_edit->text().toStdString());
            check->set_fi(ui->fi_line_edit->text().toStdString());
            check->set_operation_type(OperationType(ui->operation_type_combo_box->currentIndex() + 1));

            return check;
        } catch(OfdRequestException e) {
            QMessageBox infoDialog;
            if (!strcmp(e.what(), "Incorrect captcha")) {

                infoDialog.setText(tr("Captcha was not solved correctly!"));
                infoDialog.setIcon(QMessageBox::Critical);
                infoDialog.setWindowTitle(tr("Captcha is incorrect"));
                infoDialog.exec();
                continue;
            } else if (!strcmp(e.what(), "Internal server error")) {
                infoDialog.setText(tr("Internal server error. Please, try again later."));
                infoDialog.setIcon(QMessageBox::Critical);
                infoDialog.setWindowTitle(tr("Internal server error"));
                infoDialog.exec();
                return nullptr;
            } else if (!strcmp(e.what(), "Does not exist")) {
                infoDialog.setText(tr("Check not found. Please, ensure correctness of entered data."));
                infoDialog.setIcon(QMessageBox::Critical);
                infoDialog.setWindowTitle(tr("Check was not found"));
                infoDialog.exec();
                return nullptr;
            }
        }
    }
    return nullptr;
}

MainWindow::~MainWindow() {
    delete ui;
}

void MainWindow::on_deleteSelectedButton_clicked() {
    QItemSelectionModel *select = ui->checkQueueTable->selectionModel();
    std::vector<unsigned int> to_delete_positions = {};
    for (auto &row : select->selectedIndexes()) {
        if (row.column() != 0) continue;
        to_delete_positions.push_back(row.row());
    }
    std::sort(to_delete_positions.begin(), to_delete_positions.end(), std::greater<unsigned int>());
    for (unsigned int position : to_delete_positions) {
        model->removeRows(position, 1);
    }
    emit model->dataChanged(model->index(checks.size(), 0), model->index(checks.size() + to_delete_positions.size(), 1));
    ui->checkQueueTable->clearSelection();
}

void MainWindow::set_check_params(std::map<std::string, std::string> paramsMap) {
    ui->fn_line_edit->setText(QString::fromStdString(paramsMap["fn"]));
    ui->fd_line_edit->setText(QString::fromStdString(paramsMap["i"]));
    ui->fi_line_edit->setText(QString::fromStdString(paramsMap["fp"]));

    QString extractedDateTime = QString::fromStdString(paramsMap["t"]);
    //TODO: some QRs contain datetime in format yyyyMMddThhmmss. Perhaps there is more different formats, should write function to detect them.
    QDateTime datetime = QDateTime::fromString(extractedDateTime, "yyyyMMddThhmm");
    if (datetime == QDateTime::fromString(extractedDateTime, "20000101T1200")) {
        datetime = QDateTime::fromString(extractedDateTime, "yyyyMMddThhmmss");
    }
    ui->purchase_datetime_edit->setDateTime(datetime);

    int type = std::stoi(paramsMap["n"]);
    ui->operation_type_combo_box->setCurrentIndex(type - 1);

    std::string total = paramsMap["s"];

    ui->total_spin_box->setValue(std::stod(total));
}

